import * as tls from "tls";
import { IModbusRequest } from "./interfaces/IModbusRequest";
import { RequestFactory } from "./services/RequestFactory";
import { ConnectionOptions } from "node:tls";
import { ModbusFunctionCode } from "./constants/ModbusFunctionCode";
import { MODBUS } from "./constants/modbus.constant";
import { IModbusWriteRequest } from "./interfaces/IModbusWriteRequest";
import { ValidityCheckService } from "./services/ValidityCheckService";
import { Utils } from "./helper/Utils";

export class ModbusClient {
  private client: tls.TLSSocket;
  private responses: Map<number, any>;

  private transactionId: number;

  constructor(host: string, port: number, options: ConnectionOptions) {
    this.transactionId = 1;
    this.responses = new Map();

    this.client = tls.connect(port, host, options, () => {
      if (this.client.authorized === true) {
        console.log('Authorization worked')
      } else {
        console.log('Authorization DID NOT work')
      }
    });

    this.client.on('data', (data) => {
      console.log('on data triggered with data', data)
      const responseId = this.extractTransactionId(data);
      const promiseResolve = this.responses.get(responseId);
      if (promiseResolve) {
        promiseResolve(data);
        this.responses.delete(responseId);
      }
    });

    this.client.on('error', (error) => {
      console.log('error occured', error)

      this.client.destroy();
    });
  }

  public readCoils(modbusId: number, startRegister: number, registerCount: number): Promise<Buffer> {
    if (Utils.isInRange(registerCount, {min: 0, max: MODBUS.MAX_COILS_PER_REQUEST}) === false) {
      console.log('Invalid amount of coils. Must be in [0, 2000].')

      return;
    }

    const modbusRequest = RequestFactory.createReadRequest(
      this.transactionId,
      modbusId,
      ModbusFunctionCode.FC1,
      startRegister,
      registerCount,
    );

    console.log('Read coils going to send', modbusRequest);

    return this.sendRequest(modbusRequest);
  }
  public readDiscreteInputs(modbusId: number, startRegister: number, registerCount: number): Promise<Buffer> {
    const modbusRequest = RequestFactory.createReadRequest(
      this.transactionId,
      modbusId,
      ModbusFunctionCode.FC2,
      startRegister,
      registerCount,
    );

    console.log('Read holding register going to send', modbusRequest);

    return this.sendRequest(modbusRequest);
  }
  public readHoldingRegisters(modbusId: number, startRegister: number, registerCount: number): Promise<Buffer> {
    const modbusRequest = RequestFactory.createReadRequest(
      this.transactionId,
      modbusId,
      ModbusFunctionCode.FC3,
      startRegister,
      registerCount,
    );

    console.log('Read holding register going to send', modbusRequest);

    return this.sendRequest(modbusRequest);
  }
  public readInputRegisters(modbusId: number, startRegister: number, registerCount: number): Promise<Buffer> {
    const modbusRequest = RequestFactory.createReadRequest(
      this.transactionId,
      modbusId,
      ModbusFunctionCode.FC4,
      startRegister,
      registerCount,
    );

    console.log('Read input registers going to send', modbusRequest);

    return this.sendRequest(modbusRequest);
  }

  public writeSingleCoil(modbusId: number, startRegister: number, value: boolean): Promise<Buffer> {
    const writeValue = value === true ? MODBUS.WRITE_COIL_ENABLE_VALUE : MODBUS.WRITE_COIL_DISABLE_VALUE;
    const valueBuffer: Buffer = Buffer.alloc(MODBUS.BYTES_PER_REGISTER);
    valueBuffer.writeUint16BE(writeValue);

    const modbusRequest = RequestFactory.createWriteSingleRequest(
      this.transactionId,
      modbusId,
      ModbusFunctionCode.FC1,
      startRegister,
      valueBuffer,
    );

    console.log('Write single coil going to send', modbusRequest);

    return this.sendWriteSingleRequest(modbusRequest);
  }
  public writeSingleRegister(modbusId: number, startRegister: number, value: Buffer): Promise<Buffer> {

    const modbusRequest = RequestFactory.createWriteSingleRequest(
      this.transactionId,
      modbusId,
      ModbusFunctionCode.FC1,
      startRegister,
      value,
    );

    console.log('Write single register going to send', modbusRequest);

    return this.sendWriteSingleRequest(modbusRequest);
  }
  public writeMultipleCoils(modbusId: number, startRegister: number, values: Buffer): Promise<Buffer> {
    if (ValidityCheckService.isCoilData(values) === false) {
      console.log('Invalid coil data! Aborting request!')

      return;
    }

    const modbusRequest = RequestFactory.createWriteMultipleRequest(
      this.transactionId,
      modbusId,
      ModbusFunctionCode.FC1,
      startRegister,
      values.length,
      values,
    );

    console.log('Write multiple registers going to send', modbusRequest);

    return this.sendWriteMultipleRequest(modbusRequest);
  }
  public writeMultipleRegisters(modbusId: number, startRegister: number, values: Buffer): Promise<Buffer> {
    if (ValidityCheckService.checkRegisterCompleteness(values) === false) {
      console.log(`Error occurred while sending register with function code FC16 starting at ${startRegister}`)

      return;
    }

    const modbusRequest = RequestFactory.createWriteMultipleRequest(
      this.transactionId,
      modbusId,
      ModbusFunctionCode.FC1,
      startRegister,
      values.length / MODBUS.BYTES_PER_REGISTER,
      values,
    );

    console.log('Write multiple registers going to send', modbusRequest);

    return this.sendWriteMultipleRequest(modbusRequest);
  }

  public destroy() {
    console.log('Client will be closed now. Bye bye!');

    this.client.destroy();
  }


  /**
   * @param {IModbusRequest} modbusRequest
   * @returns {number} The requestId needed to identify the answer later
   */
  private sendRequest(modbusRequest: IModbusRequest): any {
    console.log('Sending request for', modbusRequest)
    const requestId = this.transactionId;
    const requestBuffer = RequestFactory.getRequestBuffer(modbusRequest);

    const promise = new Promise((resolve) => {
      this.responses.set(requestId, resolve);
      console.log('responses.set(requestId, resolve); DONE')
    });

    console.log('Gonna client.write')
    this.client.write(requestBuffer);
    console.log('client.write DONE')
    this.incrementTransactionId();
    console.log('transactionID is now', this.transactionId)

    return promise;
  }

  /**
   * @param {IModbusWriteRequest} modbusRequest
   * @returns {number} The requestId needed to identify the answer later
   */
  private sendWriteSingleRequest(modbusRequest: IModbusWriteRequest): any {
    console.log('Sending request for', modbusRequest)
    const requestId = this.transactionId;
    const requestBuffer = RequestFactory.getWriteSingleRequestBuffer(modbusRequest);

    const promise = new Promise((resolve) => {
      this.responses.set(requestId, resolve);
      console.log('responses.set(requestId, resolve); DONE')
    });

    console.log('Gonna client.write')
    this.client.write(requestBuffer);
    console.log('client.write DONE')
    this.incrementTransactionId();
    console.log('transactionID is now', this.transactionId)

    return promise;
  }

  /**
   * @param {IModbusWriteRequest} modbusRequest
   * @returns {number} The requestId needed to identify the answer later
   */
  private sendWriteMultipleRequest(modbusRequest: IModbusWriteRequest): any {
    console.log('Sending request for', modbusRequest)
    const requestId = this.transactionId;
    const requestBuffer = RequestFactory.getWriteMultipleRequestBuffer(modbusRequest);

    const promise = new Promise((resolve) => {
      this.responses.set(requestId, resolve);
      console.log('responses.set(requestId, resolve); DONE')
    });

    console.log('Gonna client.write')
    this.client.write(requestBuffer);
    console.log('client.write DONE')
    this.incrementTransactionId();
    console.log('transactionID is now', this.transactionId)

    return promise;
  }

  private extractTransactionId(response: Buffer): number {
    return response.readUInt16BE();
  }

  private incrementTransactionId(): void {
    this.transactionId++;

    if (this.transactionId > MODBUS.MAX_TCP_TRANSACTIONS) {
      this.transactionId = 1;
    }
  }
}