import * as tls from "tls";
import { TlsOptions } from "node:tls";
import { MODBUS } from "./constants/modbus.constant";
import { IModbusRequest } from "./interfaces/IModbusRequest";
import { RequestEvaluationService } from "./services/RequestEvaluationService";
import { ModbusFunctionCode } from "./constants/ModbusFunctionCode";
import { ModbusHelper } from "./helper/ModbusHelper";
import { Utils } from "./helper/Utils";
import { ResponseHelper } from "./helper/ResponseHelper";
import { ValidityCheckService } from "./services/ValidityCheckService";
import { AccessMode } from "./constants/AccessMode";
import { IModbusWriteRequest } from "./interfaces/IModbusWriteRequest";
import { ExceptionService } from "./services/ExceptionService";

export class ModbusServer {
  public server: tls.Server;

  private requestEvaluationService: RequestEvaluationService;
  private validityCheckService: ValidityCheckService;
  private registers: Map<ModbusFunctionCode, Map<number, number>>;
  private modbusId: number;
  private fillRegistersAllowed: boolean;

  /**
   * @param {string} host
   * @param {number} port
   * @param {TlsOptions} options Object has to contain key, cert, ca. Adding 'rejectUnauthrized: true' is highly recommended.
   */
  constructor(host: string, port: number, options: TlsOptions) {
    this.modbusId = 1;
    this.fillRegistersAllowed = true;
    this.requestEvaluationService = new RequestEvaluationService();
    this.initializeRegisterMaps();

    this.server = tls.createServer(
      options,
      (socket) => {
        socket.on('data', (data) => {
          const rawData = ModbusHelper.getRawDataString(data);
          console.log(`Received: ${rawData} [it is ${data.length} bytes long]`);

          if (this.checkMessageLength(data) === false) {
            return;
          }

          const request: IModbusRequest = this.requestEvaluationService.parseRequest(data);
          this.validityCheckService = new ValidityCheckService(request, socket, this.registers, this.modbusId, this.fillRegistersAllowed);

          if (this.validityCheckService.checkRequest() === true) {
            const functionCode = this.requestEvaluationService.getFunctionCode(data);
            console.log('Got request', request);

            switch (functionCode) {
              case ModbusFunctionCode.FC1:
              case ModbusFunctionCode.FC2:
                this.handleReadCoilRequest(request, socket);
                return;
              case ModbusFunctionCode.FC3:
              case ModbusFunctionCode.FC4:
                this.handleReadRegisterRequest(request, socket);
                return;
              case ModbusFunctionCode.FC5:
              case ModbusFunctionCode.FC15:
              case ModbusFunctionCode.FC6:
              case ModbusFunctionCode.FC16:
                const writeRequest = this.requestEvaluationService.parseWriteRequest(data);
                this.handleWriteRegisterRequest(writeRequest, socket);
                return;
              default:
                ExceptionService.sendIllegalFunctionException(request, socket);
            }
          }
        });

        socket.on('end', () => {
          console.log('Socket closed!')
        })
      },
    );

    this.server.listen(
      port,
      host,
      () => {
        console.log(`I'm listening at ${host}, on port ${port}`);
      },
    );

    this.server.on('error', (error) => {
      console.error(error);

      this.server.close();
    });
  }

  public addRegisters(
    functionCode: ModbusFunctionCode,
    startRegister: number,
    accessMode: AccessMode,
    data: Buffer,
  ): void {
    if (ValidityCheckService.checkRegisterAddress(startRegister) === false ||
        ValidityCheckService.isRegisterFunctionCode(functionCode) === false) {
      return;
    }

    if (ValidityCheckService.checkRegisterCompleteness(data) === false) {
      console.log(`Error occurred while adding registers to function code [${functionCode}] starting at ${startRegister}`)

      return;
    }

    let functionCodesToUse = ModbusHelper.getCorrespondingFunctionCodes(functionCode, accessMode);

    for (let fc of functionCodesToUse) {
      this.initRegisterMapForFunction(fc)


      const registers = ModbusHelper.bufferToRegisters(data);

      for (let i = 0; i < registers.length; i++) {
        const currentRegisters = this.registers.get(fc)

        currentRegisters.set(startRegister + i, registers[i]);
      }

      console.log(`Added ${registers.length} register to ${fc} with startAddress ${startRegister}`);
      // console.log(`number in reg 0x20: ${this.registers.get(fc).get(0x20).toString(16)}`);
      // console.log(`number in reg 0x22: ${this.registers.get(fc).get(0x21).toString(16)}`);
      // console.log(`Chars in reg 0x20: ${Utils.getCharsFromRegisterValue(this.registers.get(fc).get(0x20))}${Utils.getCharsFromRegisterValue(this.coils.get(fc).get(0x21))}`);
    }


  }

  public readRegisters(
    functionCode: ModbusFunctionCode,
    startRegister: number,
    registerCount: number,
  ) {
    if (ValidityCheckService.checkRegisterAddress(startRegister) === false ||
        ValidityCheckService.isRegisterFunctionCode(functionCode) === false) {
      return;
    }

    return this.getResponseBuffer(functionCode, startRegister, registerCount);
  }

  /**
   * Add and create have the same functionality because of the implementation of Map in Typescript.
   * Anyway it is more syntactically correct to have a method to update a register
   * @param {ModbusFunctionCode} functionCode
   * @param {number} startRegister
   * @param {AccessMode} accessMode
   * @param {Buffer} data
   */
  public setRegisters(
    functionCode: ModbusFunctionCode,
    startRegister: number,
    accessMode: AccessMode,
    data: Buffer,) {
    this.addRegisters(functionCode, startRegister, accessMode, data);
  }

  public deleteRegisters(
    functionCode: ModbusFunctionCode,
    startRegister: number,
    accessMode: AccessMode,
    registerCount: number,
  ) {
    if (ValidityCheckService.checkRegisterAddress(startRegister) === false ||
        ValidityCheckService.isRegisterFunctionCode(functionCode) === false) {
      return;
    }

    let functionCodesToUse = ModbusHelper.getCorrespondingFunctionCodes(functionCode, accessMode);

    for (let fc of functionCodesToUse) {
      for (let i = 0; i < registerCount; i++) {
        const registerToDelete = startRegister + i;

        this.registers.get(fc).delete(registerToDelete);
      }
    }
  }

  /**
   * @param {ModbusFunctionCode} functionCode
   * @param {number} startRegister
   * @param {AccessMode} accessMode
   * @param {Buffer} data Values need to be added to the buffer one byte per coil value
   */
  public addCoils(
    functionCode: ModbusFunctionCode,
    startRegister: number,
    accessMode: AccessMode,
    data: Buffer,
  ): void {
    console.log('Add coils called')
    if (ValidityCheckService.checkRegisterAddress(startRegister) === false ||
        ValidityCheckService.isCoilFunctionCode(functionCode) === false ||
        ValidityCheckService.isCoilData(data) === false) {
      console.log(`Validity check failed for adding data [${ModbusHelper.getRawDataString(data)}] to function code: ${functionCode} at register ${startRegister}`);

      return;
    }

    let functionCodesToUse = ModbusHelper.getCorrespondingFunctionCodes(functionCode, accessMode);

    for (let fc of functionCodesToUse) {
      this.initRegisterMapForFunction(fc)


      const registers = ModbusHelper.bufferToCoils(data);

      for (let i = 0; i < registers.length; i++) {
        const currentRegisters = this.registers.get(fc)

        currentRegisters.set(startRegister + i, registers[i]);
      }

      console.log(`Added ${registers.length} register to ${fc} with startAddress ${startRegister}`);
    }
  }

  public readCoils(
    functionCode: ModbusFunctionCode,
    startRegister: number,
    coilCount: number,
  ) {
    if (ValidityCheckService.checkRegisterAddress(startRegister) === false ||
        ValidityCheckService.isCoilFunctionCode(functionCode) === false) {
      return;
    }

    return this.getCoilResponseBuffer(functionCode, startRegister, coilCount);
  }


  /**
   * Add and set have the same functionality because of the implementation of Map in Typescript.
   * Anyway it is more syntactically correct to have a method to update a coil
   * @param {ModbusFunctionCode} functionCode
   * @param {number} startRegister
   * @param {AccessMode} accessMode
   * @param {Buffer} data
   */
  public setCoils(
    functionCode: ModbusFunctionCode,
    startRegister: number,
    accessMode: AccessMode,
    data: Buffer) {
    this.addCoils(functionCode, startRegister, accessMode, data);
  }

  /**
   * Same code as for registers. Added for uniformity
   * @param {ModbusFunctionCode} functionCode
   * @param {number} startRegister
   * @param {AccessMode} accessMode
   * @param {number} coilCount
   */
  public deleteCoil(
    functionCode: ModbusFunctionCode,
    startRegister: number,
    accessMode: AccessMode,
    coilCount: number,
  ) {
    this.deleteRegisters(functionCode, startRegister, accessMode, coilCount);
  }


  public setModbusId(id: number): void {
    if (Utils.isInRange(id, {min: MODBUS.ADDRESS.MIN, max: MODBUS.ADDRESS.MAX}) === false) {
      console.log('Invalid modbus id value!')

      return;
    }

    this.modbusId = id;
  }

  public enableFillRegisters(): void {
    this.fillRegistersAllowed = true;
  }

  public disableFillRegisters(): void {
    this.fillRegistersAllowed = false;
  }

  public enableFunctionCode(functionCode: ModbusFunctionCode): void {
    if (this.registers.get(functionCode) !== undefined) {
      console.log(`Function code [${functionCode}] already available`);

      return;
    }

    this.registers.set(functionCode, new Map<number, number>());
  }

  public disableFunctionCode(functionCode: ModbusFunctionCode): void {
    if (this.registers.get(functionCode) === undefined) {
      console.log(`Function code [${functionCode}] is not available!`);

      return;
    }

    this.registers.delete(functionCode);
  }

  private checkMessageLength(data: any): boolean {
    if (data.length < (MODBUS.LENGTH.MBAP_HEADER + MODBUS.LENGTH.MINIMAL_PDU)) {
      // Invalid message will be ignored
      console.log(
        'Invalid message received: ',
        ModbusHelper.getRawDataString(data),
        'Ignoring request!'
      );

      return false;
    }

    return true;
  }

  private initRegisterMapForFunction(functionCode: ModbusFunctionCode) {
    if (this.registers.get(functionCode) === undefined) {
      this.registers.set(functionCode, new Map<number, number>());
    }
  }

  private handleReadRegisterRequest(request: IModbusRequest, socket: tls.TLSSocket) {
    const responseData = this.getResponseBuffer(request.functionCode, request.registerAddress, request.registerCount);

    const responseBuffer = ResponseHelper.buildResponseBuffer(request, responseData)

    socket.write(responseBuffer)
  }

  private handleReadCoilRequest(request: IModbusRequest, socket: tls.TLSSocket) {
    try {
      const responseData = this.getCoilResponseBuffer(request.functionCode, request.registerAddress, request.registerCount);
      const responseBuffer = ResponseHelper.buildResponseBuffer(request, responseData)

      socket.write(responseBuffer)
    } catch (e) {
      ExceptionService.sendIllegalDataAddressException(request, socket, `Error while processing coil request ${e}`)

      return;
    }


  }

  private handleWriteRegisterRequest(request: IModbusWriteRequest, socket: tls.TLSSocket) {
    switch (request.functionCode) {
      case ModbusFunctionCode.FC5:
        this.handleFc5Request(request, socket);
        break;
      case ModbusFunctionCode.FC6:
        this.handleFc6Request(request, socket);
        break;
      case ModbusFunctionCode.FC15:
        this.handleFc15Request(request, socket);
        break;
      case ModbusFunctionCode.FC16:
        this.handleFc16Request(request, socket);
        break;
    }
  }

  private handleFc5Request(request: IModbusWriteRequest, socket: tls.TLSSocket) {
    if (ValidityCheckService.isSingleCoilData(request.data) === false) {
      ExceptionService.sendIllegalDataValueException(request, socket);
      return;
    }

    const value = request.data.readUInt16BE() === MODBUS.WRITE_COIL_ENABLE_VALUE ? 1 : 0;
    this.setCoils(ModbusFunctionCode.FC5, request.registerAddress, AccessMode.READ_WRITE, Buffer.from([value]));

    const responseBuffer = ResponseHelper.buildWriteSingleResponseBuffer(request)

    socket.write(responseBuffer)
  }

  private handleFc6Request(request: IModbusWriteRequest, socket: tls.TLSSocket) {
    this.setRegisters(ModbusFunctionCode.FC6, request.registerAddress, AccessMode.READ_WRITE, request.data);

    const responseBuffer = ResponseHelper.buildWriteSingleResponseBuffer(request)

    socket.write(responseBuffer)
  }

  private handleFc15Request(request: IModbusWriteRequest, socket: tls.TLSSocket) {
    const saveBuffer = this.prepareFc15Buffer(request)

    this.setCoils(ModbusFunctionCode.FC5, request.registerAddress, AccessMode.READ_WRITE, saveBuffer);

    const responseBuffer = ResponseHelper.buildWriteMultipleResponseBuffer(request);

    socket.write(responseBuffer)
  }

  private prepareFc15Buffer(request: IModbusWriteRequest): Buffer {
    const bytes = [];
    const values = [];
    let coilCount = request.registerCount;

    for (let i = 0; i < request.data.length; i++) {
      bytes.push(request.data.readUInt8(i));
    }

    for (let byte of bytes) {
      for (let i = 0; i < 8 && i < coilCount; i++) {
        values.push(Utils.getBit(byte, i));
      }

      coilCount -= 8;
    }

    return Buffer.from(values);
  }

  private handleFc16Request(request: IModbusWriteRequest, socket: tls.TLSSocket) {
    this.setRegisters(ModbusFunctionCode.FC3, request.registerAddress, AccessMode.READ_WRITE, request.data);
    this.setRegisters(ModbusFunctionCode.FC4, request.registerAddress, AccessMode.READ, request.data);

    const responseBuffer = ResponseHelper.buildWriteMultipleResponseBuffer(request);

    socket.write(responseBuffer)
  }

  private getResponseBuffer(functionCode: ModbusFunctionCode, startRegister: number, length: number): Buffer {
    console.log(`getResponseBuffer ${functionCode} ${startRegister} ${length}`)
    const responseValues = [];
    const responseBuffer = Buffer.alloc(length * MODBUS.BYTES_PER_REGISTER);

    for (let i = 0; i < length; i++) {
      const value = this.registers.get(functionCode).get(startRegister + i);

      if (value === undefined) {
        responseValues.push(MODBUS.NAN);

        continue;
      }

      responseValues.push(value);
    }
    console.log(responseValues);

    for (let i = 0; i < responseValues.length; i++) {
      console.log("now writing value", responseValues[i], 'at offset', i * MODBUS.BYTES_PER_REGISTER)
      responseBuffer.writeUInt16BE(responseValues[i], i * MODBUS.BYTES_PER_REGISTER);
    }

    return responseBuffer;
  }

  private getCoilResponseBuffer(functionCode: ModbusFunctionCode, startRegister: number, length: number): Buffer {
    console.log(`getResponseBuffer ${functionCode} ${startRegister} ${length}`)
    const responseValues = [];

    for (let i = 0; i < length; i++) {
      const value = this.registers.get(functionCode).get(startRegister + i);

      if (value === undefined) {
        throw new Error('Coil NOT defined!')
      }

      responseValues.push(value);
    }

    return ModbusHelper.convertCoilValuesToBuffer(responseValues);
  }

  private initializeRegisterMaps(): void {
    this.registers = new Map<ModbusFunctionCode, Map<number, number>>();
  }

  public close() {
    console.log('Server will be closed now. Bye bye!');

    this.server.close();
  }
}
