import { MODBUS } from "../constants/modbus.constant";
import { IModbusRequest } from "../interfaces/IModbusRequest";
import { IModbusResponse } from "../interfaces/IModbusResponse";
import { IModbusWriteRequest } from "../interfaces/IModbusWriteRequest";
import { ModbusFunctionCode } from "../constants/ModbusFunctionCode";
import { AccessMode } from "../constants/AccessMode";
import { ValidityCheckService } from "../services/ValidityCheckService";

export class ModbusHelper {
  public static getRawDataString(data: Buffer): string {
    let rawDataString: string = '';

    for (let i = 0; i < data.length; i++) {
      const byteValue = data.readUInt8(i);

      rawDataString += `${byteValue.toString(16).padStart(2, '0')} `;
    }

    return rawDataString;
  }

  public static bufferToRegisters(data: Buffer): Array<number> {
    const registers = [];

    for (let i = 0; i < data.length; i += MODBUS.BYTES_PER_REGISTER) {
      const rawValue = data.readUInt16BE(i);

      registers.push(rawValue)
    }

    return registers;
  }

  public static bufferToCoils(data: Buffer): Array<number> {
    const coils = [];

    for (let i = 0; i < data.length; i++) {
      const rawValue = data.readUInt8(i);

      coils.push(rawValue)
    }

    return coils;
  }

  public static isWriteMultipleFunction(functionCode: ModbusFunctionCode) {
    switch (functionCode) {
      case ModbusFunctionCode.FC5:
      case ModbusFunctionCode.FC6:
        return false;
      case ModbusFunctionCode.FC15:
      case ModbusFunctionCode.FC16:
        return true;
    }
  }

  public static getCorrespondingFunctionCodes(functionCode: ModbusFunctionCode, accessMode: AccessMode): Array<ModbusFunctionCode> {
    if (ValidityCheckService.isRegisterFunctionCode(functionCode) === true) {
      switch (accessMode) {
        case AccessMode.READ:
          return [functionCode];
        case AccessMode.WRITE:
          return [ModbusFunctionCode.FC6, ModbusFunctionCode.FC16]
        case AccessMode.READ_WRITE:
          return [functionCode, ModbusFunctionCode.FC6, ModbusFunctionCode.FC16]
      }
    }

    switch (accessMode) {
      case AccessMode.READ:
        return [functionCode];
      case AccessMode.WRITE:
        return [ModbusFunctionCode.FC5, ModbusFunctionCode.FC15]
      case AccessMode.READ_WRITE:
        return [functionCode, ModbusFunctionCode.FC5, ModbusFunctionCode.FC15]
    }
  }

  public static getEmptyRequest(): IModbusRequest {
    return {
      transactionId:      undefined,
      protocolIdentifier: undefined,
      messageLength:      undefined,
      modbusId:           undefined,
      functionCode:       undefined,
      registerAddress:    undefined,
      registerCount:      undefined,
    };
  }

  public static getEmptyWriteRequest(): IModbusWriteRequest {
    return {
      transactionId:      undefined,
      protocolIdentifier: undefined,
      messageLength:      undefined,
      modbusId:           undefined,
      functionCode:       undefined,
      registerAddress:    undefined,
      data:               undefined,
    };
  }

  public static getEmptyResponse(): IModbusResponse {
    return {
      transactionId:      undefined,
      protocolIdentifier: undefined,
      messageLength:      undefined,
      modbusId:           undefined,
      functionCode:       undefined,
      dataLength:         undefined,
      data:               undefined,
    };
  }

  public static convertCoilValuesToBuffer(values: Array<number>) {
    const resultBytes = [];
    let currentByte = 0;
    let bitPosition = 0;

    values.forEach((bit) => {
      currentByte |= bit << bitPosition;
      bitPosition++;

      if (bitPosition === 8 || bitPosition === values.length) {
        resultBytes.push(currentByte);
        currentByte = 0;
        bitPosition = 0;
      }
    });


    if (bitPosition > 0 && bitPosition < 8) {
      resultBytes.push(currentByte);
    }

    return Buffer.from(resultBytes);
  }

}