import { MODBUS } from "../constants/modbus.constant";
import { IModbusRequest } from "../interfaces/IModbusRequest";
import { ModbusException } from "../constants/ModbusException";
import { IModbusResponse } from "../interfaces/IModbusResponse";
import { ModbusHelper } from "./ModbusHelper";
import { IModbusWriteRequest } from "../interfaces/IModbusWriteRequest";

export class ResponseHelper {
  public static buildResponseBuffer(request: IModbusRequest, responseContent: Buffer): Buffer {
    const responseLength = responseContent.length
                           + MODBUS.LENGTH.MODBUS_ID
                           + MODBUS.LENGTH.FUNCTION_CODE
                           + MODBUS.LENGTH.DATA_LENGTH;

    const fixedResponseBuffer: Buffer = Buffer.alloc(MODBUS.LENGTH.RESPONSE.READ_WITHOUT_CONTENT);

    fixedResponseBuffer.writeUint16BE(request.transactionId, MODBUS.OFFSET.RESPONSE.TRANSACTION_ID);
    fixedResponseBuffer.writeUint16BE(MODBUS.PROTOCOL_ID, MODBUS.OFFSET.RESPONSE.PROTOCOL_ID);
    fixedResponseBuffer.writeUint16BE(responseLength, MODBUS.OFFSET.RESPONSE.MSG_LENGTH);
    fixedResponseBuffer.writeUint8(request.modbusId, MODBUS.OFFSET.RESPONSE.MODBUS_ID);
    fixedResponseBuffer.writeUint8(request.functionCode, MODBUS.OFFSET.RESPONSE.FUNCTION_CODE);
    fixedResponseBuffer.writeUint8(responseContent.length, MODBUS.OFFSET.RESPONSE.DATA_LENGTH);

    return Buffer.concat([fixedResponseBuffer, responseContent]);
  }

  public static buildWriteSingleResponseBuffer(request: IModbusWriteRequest): Buffer {
    const responseLength = MODBUS.LENGTH.WRITE_SINGLE_COIL_PDU
                           + MODBUS.LENGTH.MODBUS_ID;

    const fixedResponseBuffer: Buffer = Buffer.alloc(MODBUS.LENGTH.RESPONSE.WRITE_SINGLE_WITHOUT_CONTENT);

    fixedResponseBuffer.writeUint16BE(request.transactionId, MODBUS.OFFSET.WRITE_SINGLE.TRANSACTION_ID);
    fixedResponseBuffer.writeUint16BE(MODBUS.PROTOCOL_ID, MODBUS.OFFSET.WRITE_SINGLE.PROTOCOL_ID);
    fixedResponseBuffer.writeUint16BE(responseLength, MODBUS.OFFSET.WRITE_SINGLE.MSG_LENGTH);
    fixedResponseBuffer.writeUint8(request.modbusId, MODBUS.OFFSET.WRITE_SINGLE.MODBUS_ID);
    fixedResponseBuffer.writeUint8(request.functionCode, MODBUS.OFFSET.WRITE_SINGLE.FUNCTION_CODE);
    fixedResponseBuffer.writeUint16BE(request.registerAddress, MODBUS.OFFSET.WRITE_SINGLE.REG_ADDRESS);

    return Buffer.concat([fixedResponseBuffer, request.data]);
  }


  public static buildWriteMultipleResponseBuffer(request: IModbusWriteRequest): Buffer {
    const responseLength = MODBUS.LENGTH.WRITE_MULTIPLE_REGISTERS_RESPONSE_PDU
                           + MODBUS.LENGTH.MODBUS_ID;

    const responseBuffer: Buffer = Buffer.alloc(MODBUS.LENGTH.RESPONSE.WRITE_MULTIPLE);

    responseBuffer.writeUint16BE(request.transactionId, MODBUS.OFFSET.WRITE_MULTIPLE.TRANSACTION_ID);
    responseBuffer.writeUint16BE(MODBUS.PROTOCOL_ID, MODBUS.OFFSET.WRITE_MULTIPLE.PROTOCOL_ID);
    responseBuffer.writeUint16BE(responseLength, MODBUS.OFFSET.WRITE_MULTIPLE.MSG_LENGTH);
    responseBuffer.writeUint8(request.modbusId, MODBUS.OFFSET.WRITE_MULTIPLE.MODBUS_ID);
    responseBuffer.writeUint8(request.functionCode, MODBUS.OFFSET.WRITE_MULTIPLE.FUNCTION_CODE);
    responseBuffer.writeUint16BE(request.registerAddress, MODBUS.OFFSET.WRITE_MULTIPLE.REG_ADDRESS);
    responseBuffer.writeUint16BE(request.registerCount, MODBUS.OFFSET.WRITE_MULTIPLE.REG_COUNT);

    return responseBuffer;
  }


  public static buildExceptionResponseBuffer(request: IModbusRequest | IModbusWriteRequest, exception: ModbusException): Buffer {
    const responseBuffer: Buffer = Buffer.alloc(MODBUS.LENGTH.EXCEPTION);

    responseBuffer.writeUint16BE(request.transactionId, MODBUS.OFFSET.RESPONSE.TRANSACTION_ID);
    responseBuffer.writeUint16BE(MODBUS.PROTOCOL_ID, MODBUS.OFFSET.RESPONSE.PROTOCOL_ID);
    responseBuffer.writeUint16BE(MODBUS.LENGTH.RESPONSE.READ_WITH_EXCEPTION, MODBUS.OFFSET.RESPONSE.MSG_LENGTH);
    responseBuffer.writeUint8(request.modbusId, MODBUS.OFFSET.RESPONSE.MODBUS_ID);
    responseBuffer.writeUint8(request.functionCode + MODBUS.EXCEPTION_FUNCTION_CODE_OFFSET, MODBUS.OFFSET.RESPONSE.FUNCTION_CODE);
    responseBuffer.writeUint8(exception, MODBUS.OFFSET.RESPONSE.EXCEPTION);

    return responseBuffer;
  }

  public static extractResponse(response: Buffer): IModbusResponse {
    const modbusResponse = ModbusHelper.getEmptyResponse();

    modbusResponse.transactionId = response.readUInt16BE(MODBUS.OFFSET.RESPONSE.TRANSACTION_ID);
    modbusResponse.protocolIdentifier = response.readUInt16BE(MODBUS.OFFSET.RESPONSE.PROTOCOL_ID);
    modbusResponse.messageLength = response.readUInt16BE(MODBUS.OFFSET.RESPONSE.MSG_LENGTH);
    modbusResponse.modbusId = response.readUInt8(MODBUS.OFFSET.RESPONSE.MODBUS_ID);
    modbusResponse.functionCode = response.readUInt8(MODBUS.OFFSET.RESPONSE.FUNCTION_CODE);

    if (modbusResponse.functionCode > MODBUS.EXCEPTION_FUNCTION_CODE_OFFSET) {
      // Response is a Modbus Exception code. Data will contain ModbusException
      modbusResponse.dataLength = 1;
      modbusResponse.data = response.subarray(MODBUS.OFFSET.RESPONSE.EXCEPTION)

      const exceptionValue = modbusResponse.data.readUInt8()
      try {
        modbusResponse.exception = exceptionValue as ModbusException;
      } catch (error) {
        console.log('Invalid Modbus Exception code', exceptionValue);
        modbusResponse.exception = ModbusException.NotModbus;
      }

      return modbusResponse;
    }

    modbusResponse.dataLength = response.readUInt8(MODBUS.OFFSET.RESPONSE.DATA_LENGTH);
    modbusResponse.data = response.subarray(MODBUS.OFFSET.RESPONSE.DATA)

    return modbusResponse;
  }
}