import { MODBUS } from "../constants/modbus.constant";
import { IModbusRequest } from "../interfaces/IModbusRequest";
import { ModbusError } from "../constants/ModbusError";
import { ModbusException } from "../constants/ModbusException";
import { ModbusFunctionCode } from "../constants/ModbusFunctionCode";
import { Utils } from "../helper/Utils";
import { ModbusHelper } from "../helper/ModbusHelper";
import { IModbusWriteRequest } from "../interfaces/IModbusWriteRequest";

export class RequestEvaluationService {
  public parseRequest(data: any): IModbusRequest {
    let request: IModbusRequest = ModbusHelper.getEmptyRequest();

    try {
      request.transactionId = this.getTransactionId(data);
      request.protocolIdentifier = this.getProtocolIdentifier(data);
      request.messageLength = this.getMessageLength(data);
      request.modbusId = this.getModbusId(data);
      request.functionCode = this.getFunctionCode(data);
      request.registerAddress = this.getRegisterAddress(data);
      request.registerCount = this.getRegisterCount(data);
    } catch (e) {
      console.log(`Error occured while parsing request. Modbus Exception: ${e?.exception}`, e);
    }

    return request;
  }

  public parseWriteRequest(data: any): IModbusWriteRequest {
    let request: IModbusWriteRequest = ModbusHelper.getEmptyWriteRequest();

    try {
      request.transactionId = this.getTransactionId(data);
      request.protocolIdentifier = this.getProtocolIdentifier(data);
      request.messageLength = this.getMessageLength(data);
      request.modbusId = this.getModbusId(data);
      request.functionCode = this.getFunctionCode(data);
      request.registerAddress = this.getRegisterAddress(data);
      request.data = this.getWriteValues(data, request.functionCode);

      if (ModbusHelper.isWriteMultipleFunction(request.functionCode) === true) {
        request.registerCount = this.getWriteCount(data);
        request.byteCount = this.getByteCount(data);
      }

    } catch (e) {
      console.log(`Error occured while parsing request. Modbus Exception: ${e?.exception}`, e);
    }

    return request;
  }

  public getFunctionCode(data: any): ModbusFunctionCode {
    const value = data.readUInt8(MODBUS.OFFSET.REQUEST.FUNCTION_CODE);

    try {
      return value as ModbusFunctionCode;
    } catch (e) {
      throw new ModbusError(ModbusException.IllegalFunction)
    }
  }

  private getTransactionId(data: any): number {
    return data.readUInt16BE(MODBUS.OFFSET.REQUEST.TRANSACTION_ID);
  }

  private getProtocolIdentifier(data: any): number {
    const value = data.readUInt16BE(MODBUS.OFFSET.REQUEST.PROTOCOL_ID);

    if (value !== MODBUS.PROTOCOL_ID) {
      throw new ModbusError(ModbusException.NotModbus)
    }

    return value;
  }

  private getMessageLength(data: any): number {
    const value = data.readUInt16BE(MODBUS.OFFSET.REQUEST.MSG_LENGTH);

    console.log('getMsgLength value', value)

    if (value < MODBUS.LENGTH.READ_REGISTER_REQUEST_PDU) {
      throw new ModbusError(ModbusException.NotModbus)
    }

    return value;
  }

  private getModbusId(data: any): number {
    const value = data.readUInt8(MODBUS.OFFSET.REQUEST.MODBUS_ID);

    if (value < MODBUS.ADDRESS.MIN || value > MODBUS.ADDRESS.MAX) {
      throw new ModbusError(ModbusException.IllegalDataAddress)
    }

    return value;
  }

  private getRegisterAddress(data: any): number {
    return data.readUInt16BE(MODBUS.OFFSET.REQUEST.REG_ADDRESS);
  }

  private getRegisterCount(data: any): number {
    const value = data.readUInt16BE(MODBUS.OFFSET.REQUEST.REG_COUNT);

    if (Utils.isInRange(value, MODBUS.READ_REGISTER_COUNT_RANGE) === false) {
      throw new ModbusError(ModbusException.IllegalDataAddress)
    }

    return value;
  }

  private getWriteValues(data: any, functionCode: ModbusFunctionCode): Buffer {
    let index = ModbusHelper.isWriteMultipleFunction(functionCode) === true ?
      MODBUS.OFFSET.WRITE_MULTIPLE.DATA :
      MODBUS.OFFSET.WRITE_SINGLE.DATA;

    return (data as Buffer).subarray(index);
  }


  private getWriteCount(data: any): number {
    const value = data.readUInt16BE(MODBUS.OFFSET.WRITE_MULTIPLE.REG_COUNT);

    if (Utils.isInRange(value, MODBUS.WRITE_REGISTER_COUNT_RANGE) === false) {
      throw new ModbusError(ModbusException.IllegalDataValue)
    }

    return value;
  }


  private getByteCount(data: any): number {
    return data.readUInt8(MODBUS.OFFSET.WRITE_MULTIPLE.BYTE_COUNT);
  }
}