import { IModbusRequest } from "../interfaces/IModbusRequest";
import { MODBUS } from "../constants/modbus.constant";
import { ModbusFunctionCode } from "../constants/ModbusFunctionCode";
import { Utils } from "../helper/Utils";
import { ModbusHelper } from "../helper/ModbusHelper";
import { IModbusWriteRequest } from "../interfaces/IModbusWriteRequest";

export class RequestFactory {
  /**
   * @param {number} transactionId
   * @param {number} modbusId
   * @param {ModbusFunctionCode} functionCode
   * @param {number} registerAddress
   * @param {number} registerCount
   * @returns {IModbusRequest} ModbusRequest with given parameters or empty ModbusRequest if parameter was invalid
   */
  public static createReadRequest(
    transactionId: number,
    modbusId: number,
    functionCode: ModbusFunctionCode,
    registerAddress: number,
    registerCount: number,
  ): IModbusRequest {
    if (
      Utils.isInRange(transactionId, {min: 1, max: MODBUS.MAX_TCP_TRANSACTIONS}) === false ||
      Utils.isInRange(modbusId, {min: MODBUS.ADDRESS.BROADCAST, max: MODBUS.ADDRESS.MAX}) === false ||
      Utils.isInRange(registerAddress, {min: 0, max: 0xFFFF}) === false ||
      Utils.isInRange(registerCount, {min: 1, max: MODBUS.MAX_REGISTER_PER_REQUEST}) === false
    ) {
      console.log('Invalid request parameter!');

      return ModbusHelper.getEmptyRequest();
    }

    return {
      transactionId,
      protocolIdentifier: MODBUS.PROTOCOL_ID,
      messageLength: MODBUS.LENGTH.REQUEST_MESSAGE,
      modbusId,
      functionCode,
      registerAddress,
      registerCount,
    }
  }

  public static getRequestBuffer(modbusRequest: IModbusRequest): Buffer {
    const requestBuffer = Buffer.alloc(MODBUS.LENGTH.MBAP_HEADER + MODBUS.LENGTH.READ_REGISTER_REQUEST_PDU);

    requestBuffer.writeUint16BE(modbusRequest.transactionId, MODBUS.OFFSET.REQUEST.TRANSACTION_ID)
    requestBuffer.writeUint16BE(modbusRequest.protocolIdentifier, MODBUS.OFFSET.REQUEST.PROTOCOL_ID)
    requestBuffer.writeUint16BE(modbusRequest.messageLength, MODBUS.OFFSET.REQUEST.MSG_LENGTH)
    requestBuffer.writeUint8(modbusRequest.modbusId, MODBUS.OFFSET.REQUEST.MODBUS_ID)
    requestBuffer.writeUint8(modbusRequest.functionCode, MODBUS.OFFSET.REQUEST.FUNCTION_CODE)
    requestBuffer.writeUint16BE(modbusRequest.registerAddress, MODBUS.OFFSET.REQUEST.REG_ADDRESS)
    requestBuffer.writeUint16BE(modbusRequest.registerCount, MODBUS.OFFSET.REQUEST.REG_COUNT)

    return requestBuffer;
  }

  /**
   * @param {number} transactionId
   * @param {number} modbusId
   * @param {ModbusFunctionCode} functionCode
   * @param {number} registerAddress
   * @param {Buffer} value
   * @returns {IModbusRequest} ModbusRequest with given parameters or empty ModbusRequest if parameter was invalid
   */
  public static createWriteSingleRequest(
    transactionId: number,
    modbusId: number,
    functionCode: ModbusFunctionCode,
    registerAddress: number,
    value: Buffer,
  ): IModbusWriteRequest {
    if (
      Utils.isInRange(transactionId, {min: 1, max: MODBUS.MAX_TCP_TRANSACTIONS}) === false ||
      Utils.isInRange(modbusId, {min: MODBUS.ADDRESS.BROADCAST, max: MODBUS.ADDRESS.MAX}) === false ||
      Utils.isInRange(registerAddress, {min: 0, max: 0xFFFF}) === false ||
      value.length !== MODBUS.BYTES_PER_REGISTER
    ) {
      console.log('Invalid request parameter!');

      return ModbusHelper.getEmptyWriteRequest();
    }

    return {
      transactionId,
      protocolIdentifier: MODBUS.PROTOCOL_ID,
      messageLength: MODBUS.LENGTH.REQUEST_MESSAGE,
      modbusId,
      functionCode,
      registerAddress,
      data: value,
    }
  }


  public static getWriteSingleRequestBuffer(modbusRequest: IModbusWriteRequest): Buffer {
    const requestBuffer = Buffer.alloc(MODBUS.LENGTH.MBAP_HEADER + MODBUS.LENGTH.WRITE_SINGLE_REGISTER_PDU);

    const offsets = MODBUS.OFFSET.WRITE_SINGLE;

    requestBuffer.writeUint16BE(modbusRequest.transactionId, offsets.TRANSACTION_ID)
    requestBuffer.writeUint16BE(modbusRequest.protocolIdentifier, offsets.PROTOCOL_ID)
    requestBuffer.writeUint16BE(modbusRequest.messageLength, offsets.MSG_LENGTH)
    requestBuffer.writeUint8(modbusRequest.modbusId, offsets.MODBUS_ID)
    requestBuffer.writeUint8(modbusRequest.functionCode, offsets.FUNCTION_CODE)
    requestBuffer.writeUint16BE(modbusRequest.registerAddress, offsets.REG_ADDRESS)
    requestBuffer.writeUint16BE(modbusRequest.data.readUInt16BE(), offsets.DATA)

    return requestBuffer;
  }


  /**
   * @param {number} transactionId
   * @param {number} modbusId
   * @param {ModbusFunctionCode} functionCode
   * @param {number} registerAddress
   * @param {number} count
   * @param {Buffer} values
   * @returns {IModbusRequest} ModbusRequest with given parameters or empty ModbusRequest if parameter was invalid
   */
  public static createWriteMultipleRequest(
    transactionId: number,
    modbusId: number,
    functionCode: ModbusFunctionCode,
    registerAddress: number,
    count: number,
    values: Buffer,
  ): IModbusWriteRequest {
    if (
      Utils.isInRange(transactionId, {min: 1, max: MODBUS.MAX_TCP_TRANSACTIONS}) === false ||
      Utils.isInRange(modbusId, {min: MODBUS.ADDRESS.BROADCAST, max: MODBUS.ADDRESS.MAX}) === false ||
      Utils.isInRange(registerAddress, {min: 0, max: 0xFFFF}) === false ||
      values.length < MODBUS.BYTES_PER_REGISTER ||
      ModbusHelper.isWriteMultipleFunction(functionCode) === false
    ) {
      console.log('Invalid request parameter!');

      return ModbusHelper.getEmptyWriteRequest();
    }

    const getByteCount = functionCode === ModbusFunctionCode.FC15 ?
      (valueCount) => { return Math.ceil(valueCount / 8)} :
      (valueCount) => { return valueCount * MODBUS.BYTES_PER_REGISTER}
    ;

    return {
      transactionId,
      protocolIdentifier: MODBUS.PROTOCOL_ID,
      messageLength: MODBUS.LENGTH.REQUEST_MESSAGE,
      modbusId,
      functionCode,
      registerAddress,
      registerCount: count,
      byteCount: getByteCount(count),
      data: values,
    }
  }


  public static getWriteMultipleRequestBuffer(modbusRequest: IModbusWriteRequest): Buffer {
    const requestBuffer = Buffer.alloc(MODBUS.LENGTH.MBAP_HEADER + MODBUS.LENGTH.WRITE_MULTIPLE_REGISTERS_REQUEST_PDU_WITHOUT_PAYLOAD);

    const offsets = MODBUS.OFFSET.WRITE_MULTIPLE;

    requestBuffer.writeUint16BE(modbusRequest.transactionId, offsets.TRANSACTION_ID)
    requestBuffer.writeUint16BE(modbusRequest.protocolIdentifier, offsets.PROTOCOL_ID)
    requestBuffer.writeUint16BE(modbusRequest.messageLength, offsets.MSG_LENGTH)
    requestBuffer.writeUint8(modbusRequest.modbusId, offsets.MODBUS_ID)
    requestBuffer.writeUint8(modbusRequest.functionCode, offsets.FUNCTION_CODE)
    requestBuffer.writeUint16BE(modbusRequest.registerAddress, offsets.REG_ADDRESS)
    requestBuffer.writeUint16BE(modbusRequest.registerCount, offsets.REG_COUNT)
    requestBuffer.writeUint8(modbusRequest.byteCount, offsets.BYTE_COUNT)

    if (modbusRequest.functionCode === ModbusFunctionCode.FC15) {
      const coilValues = [];

      for (let i = 0; i < modbusRequest.data.length; i++) {
        coilValues.push(modbusRequest.data.readUInt8(i))
      }

      modbusRequest.data = ModbusHelper.convertCoilValuesToBuffer(coilValues);
    }

    return Buffer.concat([requestBuffer, modbusRequest.data]);
  }


}