import { IModbusRequest } from "../interfaces/IModbusRequest";
import * as tls from "tls";
import { ResponseHelper } from "../helper/ResponseHelper";
import { ModbusException } from "../constants/ModbusException";
import { ModbusFunctionCode } from "../constants/ModbusFunctionCode";
import { Utils } from "../helper/Utils";
import { MODBUS } from "../constants/modbus.constant";

export class ValidityCheckService {
  private modbusId: number;
  private allowFillRegisters: boolean;
  private request: IModbusRequest;
  private socket: tls.TLSSocket;
  private registers: Map<ModbusFunctionCode, Map<number, number>>;


  public constructor(
    request: IModbusRequest,
    socket: tls.TLSSocket,
    registers: Map<ModbusFunctionCode, Map<number, number>>,
    modbusId: number,
    allowFillRegisters: boolean,
  ) {
    this.request = request;
    this.socket = socket;
    this.registers = registers;
    this.modbusId = modbusId;
    this.allowFillRegisters = allowFillRegisters;
  }

  public static checkRegisterAddress(startRegister: number): boolean {
    if (Utils.isInRange(startRegister, MODBUS.REGISTER_RANGE) === false) {
      console.log(`Register [${startRegister}] out of range!`)

      return false;
    }

    return true;
  }


  public static checkRegisterCompleteness(data: Buffer): boolean {
    if (data.length % MODBUS.BYTES_PER_REGISTER !== 0) {
      console.log(`Buffer contains an incomplete register!!`)

      return false;
    }

    return true;
  }

  public static isRegisterFunctionCode(functionCode: ModbusFunctionCode) {
    if (MODBUS.REGISTER_FUNCTION_CODES.includes(functionCode) === false) {
      console.log('Not a register based function code. Use `addCoils()` instead!');

      return false;
    }

    return true;
  }

  public static isCoilFunctionCode(functionCode: ModbusFunctionCode) {
    if (MODBUS.COIL_FUNCTION_CODES.includes(functionCode) === false) {
      console.log('Not a coil based function code. Use `addRegisters()` instead!');

      return false;
    }

    return true;
  }

  public static isCoilData(data: Buffer) {
    for (let i = 0; i < data.length; i++) {
      const byteValue = data.readUInt8(i);

      if (Utils.isInRange(byteValue, MODBUS.COIL_VALUE_RANGE) === false) {
        console.log('Trying to add invalid coil value:', byteValue)

        return false;
      }
    }

    return true;
  }

  public static isSingleCoilData(data: Buffer) {
    const value = data.readUInt16BE();

    if (value !== MODBUS.WRITE_COIL_ENABLE_VALUE && value !== MODBUS.WRITE_COIL_DISABLE_VALUE) {
      return false;
    }

    return true;
  }

  public checkRequest(): boolean {
    return this.checkFunctionCode() && this.checkModbusId() && this.checkRegisterAvailability();
  }

  private checkFunctionCode(): boolean {
    if (this.registers.get(this.request.functionCode) === undefined) {
      // No data available for this function code -> ModbusException IllegalFunction
      console.log('Request with invalid function code:', this.request.functionCode);

      const exceptionResponseBuffer = ResponseHelper.buildExceptionResponseBuffer(this.request, ModbusException.IllegalFunction);
      this.socket.write(exceptionResponseBuffer);

      return false;
    }

    return true;
  }

  private checkModbusId(): boolean {
    if (this.request.modbusId !== this.modbusId) {
      // Request for other slave, will be ignored
      console.log('Request for other slave received. This request will be ignored!')

      return false;
    }

    return true;
  }

  /**
   * Checks whether requested start register is available and, if fill registers are not allowed, whether all requested
   * values are available.
   * @returns {boolean}
   */
  private checkRegisterAvailability(): boolean {
    const startAddress = this.request.registerAddress;
    const lastRegister = startAddress + this.request.registerCount -1;
    const availableRegisters = this.registers.get(this.request.functionCode);

    if (availableRegisters.get(startAddress) === undefined) {
      return false;
    }

    for (let i = (startAddress + 1); i <= lastRegister; i++) {
      if (availableRegisters.get(i) === undefined && this.allowFillRegisters === false) {
        return false;
      }
    }

    return true;
  }
}